###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.database.ufo_base import UfoBase
from onyx.database.objdb import ObjNotFound
from onyx.database.objdb_api import GetObj, DelObj
from onyx.depgraph.graph import DependencyGraph, CreateNode, GraphNodeSettable
from onyx.depgraph.graph import GraphNodeError

import onyx.database as onyx_db
import onyx.depgraph as onyx_dg

import functools

__all__ = [
    "UseGraph",
    "GraphNodeVt",
    "CreateInMemory",
    "RemoveFromGraph",
    "PurgeObj",
    "GetNode",
    "GetVal",
    "SetVal",
    "InvalidateNode",
    "NodesByObjName",
    "ValueTypesByInstance",
    "ChildrenSet",
    "LeafNodes",
    "SettableNodes",
]


###############################################################################
class UseGraph(object):
    """
    Context manager used to activate the dependency graph
    """
    def __init__(self, graph=None):
        self.graph = graph or DependencyGraph()

    def __enter__(self):
        self.current = onyx_dg.graph
        onyx_dg.graph = self.graph

    def __exit__(self, *args):
        onyx_dg.graph.clear()
        onyx_dg.graph = self.current
        # --- returns False so that all execptions raised will be propagated
        return False


###############################################################################
#  Base class for ValueTypes
class ValueType(object):
    pass


###############################################################################
class CalcVT(ValueType):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), "vt") works fine
            return self

        self.instance = instance
        return self

    def __set__(self, instance, value):
        raise AttributeError("Cannot set a Calc VT")

    def __call__(self, *args, **kwds):
        caller = (self.instance.Name, self.func.__name__)
        return self.func(self.instance,
                         functools.partial(GetNodeVal, caller), *args, **kwds)


###############################################################################
class SettableVT(ValueType):
    def __init__(self, fget, fset, name):
        # --- make sure that getter and setter functions have the following
        #     signature:
        #         fget(self, graph), fset(self, graph, value)
        #
        if fget.__code__.co_argcount < 2:
            raise GraphNodeError("Missing self and/or "
                                 "graph in the getter definition")
        if fget.__code__.co_argcount > 2:
            raise GraphNodeError("A Settable VT getter cannot take "
                                 "extra arguments besides self and graph")
        if fset.__code__.co_argcount < 3:
            raise GraphNodeError("Missing self and/or graph "
                                 "and/or value in the setter definition")
        if fset.__code__.co_argcount > 3:
            raise GraphNodeError("A Settable VT setter cannot take extra "
                                 "arguments besides self, graph, and value")
        self.fget = fget
        self.fset = fset
        self.__name__ = name

    def __get__(self, instance, owner=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), "vt") works fine
            return self

        caller = (instance.Name, self.__name__)
        return self.fget(instance, functools.partial(GetNodeVal, caller))

    def __set__(self, instance, value):
        caller = (instance.Name, self.__name__)
        self.fset(instance, functools.partial(GetNodeVal, caller), value)


###############################################################################
class PropertyVT(ValueType):
    def __init__(self, fget):
        # --- make sure that the decorated method is of the type:
        #         f(self, graph)
        #
        if fget.__code__.co_argcount < 2:
            raise GraphNodeError("Missing self and/or "
                                 "graph in the method definition")
        if fget.__code__.co_argcount > 2:
            raise GraphNodeError("A Property VT cannot take "
                                 "extra arguments besides self and graph")
        self.fget = fget

    def __get__(self, instance, owner=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), "vt") works fine
            return self

        caller = (instance.Name, self.fget.__name__)
        return self.fget(instance, functools.partial(GetNodeVal, caller))

    def __set__(self, instance, value):
        raise AttributeError("Cannot set a Property VT")


###############################################################################
class GraphNodeVt(object):
    """
    Calss-method decorator for objects derived from UfoBase.
    """
    # -------------------------------------------------------------------------
    def __init__(self, vt_type="Calc", fget=None, fset=None):
        self.vt_type = vt_type
        if vt_type == "Settable":
            if fget is None or fset is None:
                raise GraphNodeError("Settable VT needs "
                                     "both getter and setter")
            else:
                self.fget = fget
                self.fset = fset

    # -------------------------------------------------------------------------
    def __call__(self, func):
        if self.vt_type == "Calc":
            return CalcVT(func)
        elif self.vt_type == "Property":
            return PropertyVT(func)
        elif self.vt_type == "Settable":
            return SettableVT(self.fget, self.fset, func.__name__)
        else:
            GraphNodeError("Unrecognized VT type {0:s}".format(self.vt_type))


# -----------------------------------------------------------------------------
def CreateInMemory(instance):
    """
    Description:
        Helper function that adds an instance of an object derived from
        UfoBase in memory so that  all its VTs are on-graph and their values
        can be retrieved using the standard API (GetVal, GetNodeValue or graph
        when from within a mothod decorated by GraphNodeVt).
    Inputs:
        instance - the instance of an object derived from UfoBase
    Returns:
        The instance itself.
    """
    try:
        # --- return a reference to the cached instance with same name
        return onyx_dg.graph.objects[instance.Name]
    except KeyError:
        # --- add object instance to graph cache
        onyx_dg.graph.objects[instance.Name] = instance
        # --- add object instance to database weak references
        onyx_db.obj_clt.refs[instance.Name] = instance
        # --- return a reference to the instance
        return instance


# -----------------------------------------------------------------------------
def RemoveFromGraph(name):
    """
    Description:
        Remove all VTs of an object from the graph.
    Inputs:
        name - the name of an instance of an object derived from UfoBase
    Returns:
        None.
    """
    # --- make sure that name is indeed the object's Name
    if isinstance(name, UfoBase):
        name = name.Name

    # --- first remove from the graph all nodes referring to this object
    for node in NodesByObjName(name):
        # --- reconstruct node name
        node_name = (node.obj_ref().Name, node.VT)
        # --- invalidate parents and remove node from children list
        for parent in node.parents:
            onyx_dg.graph[parent].invalidate()
            onyx_dg.graph[parent].children.remove(node_name)
        # --- remove from parents list of each child
        for child in node.children:
            onyx_dg.graph[child].parents.remove(node_name)
        # --- remove node from the graph
        del onyx_dg.graph[node_name]

#    # --- remove the object from the database cache (if present)
#    # FIXME: are we sure we need this???
#    del onyx_db.obj_clt[name]

    # --- finally remove the object itself from the graph's cache and
    #     VT mapping (if present)
    onyx_dg.graph.vts_by_obj.pop(name, None)
    onyx_dg.graph.objects.pop(name, None)


# -----------------------------------------------------------------------------
def PurgeObj(obj):
    """
    Description:
        Remove all VTs of an object from the graph and then delete it from
        the backend.
    Inputs:
        obj - an instance of an object derived from UfoBase (or its name)
    Returns:
        None.
    """
    RemoveFromGraph(obj)
    try:
        DelObj(obj)
    except ObjNotFound:
        pass


# -----------------------------------------------------------------------------
def GetNode(target):
    """
    Description:
        Helper function: Gets a GraphNode from the graph or creates it, if not
        present.
    Inputs:
        target - a (Object, VT) tuple
    Returns:
        A hard reference to the node in the graph.
    """
    try:
        return onyx_dg.graph[target]
    except KeyError:
        onyx_dg.graph[target] = node = CreateNode(target)
        return node


# -----------------------------------------------------------------------------
def GetNodeVal(caller, obj, VT, *args, **kwds):
    """
    Description:
        This is the proper way of calling a VT on-graph. Use this method to
        get a value type from within a method that has been decorated by
        @GraphNodeVt().
    Inputs:
        caller - caller node identity
        obj    - target object as defined in database (or memory)
        VT     - target value type (a stored attribute or a method of a ufo
                 class decorated by @GraphNodeVt())
        *args  - positional arguments used to call the target method
        **kwds -      named arguments used to call the target method
    Returns:
        The VT value.
    """
    # --- get the calling node from the graph (the node is created if it
    #     doesn't exsist)
    caller_node = GetNode(caller)

    # --- define target tuple (equivalent to the calling vti)
    if isinstance(obj, str):
        target = (obj, VT)
    else:
        target = (obj.Name, VT)

    # --- add target node name to its Children
    caller_node.children.add(target)
    # --- get the target node from the graph (the node is created if it
    #     doesn't exsist)
    target_node = GetNode(target)
    # --- add calling node to its Parents
    target_node.parents.add(caller)
    # --- now that the graph is set up, return target node value
    return target_node.get_value(*args, **kwds)


# -----------------------------------------------------------------------------
def GetVal(obj, VT, *args, **kwds):
    """
    Description:
        This is the proper way of calling a VT off-graph. Use this method to
        get a value type from a script.
    Inputs:
        obj    - target object in database (or memory)
        VT     - target value type (a stored attribute or a method of a ufo
                 class decorated by @GraphNodeVt())
        *args  - positional arguments used to call the target method
        **kwds -      named arguments used to call the target method
    Returns:
        The VT value.
    """
    # --- make sure that obj is the object's Name
    if isinstance(obj, UfoBase):
        obj = obj.Name
    # --- return node value
    return GetNode((obj, VT)).get_value(*args, **kwds)


# -----------------------------------------------------------------------------
def SetVal(obj, VT, value):
    """
    Description:
        Set the value of a stored attribute on-graph (i.e. the new value is
        visible on the graph). When the instance is persisted by calling
        UpdateObj(obj.Name), the new attribute's value is persisted as well.
        NB: only leaf nodes (i.e. nodes that are object's stored attributes)
            can be set.
    Inputs:
        obj   - target object in database (or memory), or its name
        VT    - target value type (a stored attribute or a method of a ufo
                class decorated by @GraphNodeVt())
        value - the value to set the VT to
    Returns:
        None.
    """
    # --- make sure that obj is the object's Name
    if isinstance(obj, UfoBase):
        obj = obj.Name

    # --- get the target node from the graph
    name = (obj, VT)
    node = GetNode(name)

    if not isinstance(node, GraphNodeSettable):
        raise GraphNodeError("({0:s},{1:s}) "
                             "is not a settable VT".format(*name))

    # --- remove this node from the set of parents of all its children
    for child in node.children:
        onyx_dg.graph[child].parents.remove(name)

    # --- remove all its children
    node.children = set()

    # --- invalidate all its parents
    for parent in node.parents:
        onyx_dg.graph[parent].invalidate()

    # --- set the node value and set its state to valid
    node.value = value
    node.valid = True

    # --- set attribute in the cached instance of the object
    setattr(node.obj_ref(), node.VT, node.value)


# -----------------------------------------------------------------------------
def InvalidateNode(obj, VT):
    """
    Description:
        Invalidate a node (defined by the tuple (obj, VT)) in the graph
        (if present).
    Inputs:
        obj - target object in database (or memory)
        VT  - target value type (a stored attribute or a method of a ufo class
                 decorated by @GraphNodeVt())
    Returns:
        None.
    """
    if isinstance(obj, UfoBase):
        obj = obj.Name
    GetNode((obj, VT)).invalidate()


# -----------------------------------------------------------------------------
def NodesByObjName(name):
    """
    Description:
        Generator: returns all the nodes that reference the same object
    Inputs:
        name - target object's name in database (or memory)
    Yields:
        Graph nodes.
    """
    try:
        for VT in onyx_dg.graph.vts_by_obj[name]:
            yield onyx_dg.graph[(name, VT)]
    except KeyError:
        raise StopIteration


# -----------------------------------------------------------------------------
def ValueTypesByInstance(obj):
    """
    Description:
        Returns a list of all value types for object obj
    Inputs:
        obj - target object in database (or memory)
    Returns:
        A list of VT names.
    """
    if isinstance(obj, str):
        obj = GetObj(obj)

    cls = obj.__class__
    calc = [name for name, VT in
            cls.__dict__.items() if isinstance(VT, ValueType)]

    return list(obj.StoredAttrs) + calc


# -----------------------------------------------------------------------------
def children_iterator(obj_name, VT, child_VT, obj_type):
    for name, vt in onyx_dg.graph[(obj_name, VT)].children:
        if vt == child_VT:
            if obj_type is None:
                yield name
            else:
                obj = onyx_dg.graph.objects[name]
                bases = [obj.__class__.__name__]
                bases += [cls.__name__ for cls in obj.__class__.__bases__]
                if obj_type in bases:
                    yield name
        else:
            for kid in children_iterator(name, vt, child_VT, obj_type):
                yield kid


# -----------------------------------------------------------------------------
def ChildrenSet(obj, VT, child_VT, obj_type=None):
    """
    Description:
        Return all the children of a given node with a specific VT.
    Inputs:
        obj      - object in database (or memory)
        VT       - parent value type
        child_VT - children value type
        obj_type - if set, children must be subclass of this class
    Yields:
        A set with object names.
    """
    if not isinstance(obj, str):
        obj = obj.Name

    # FIXME: have to evaluate this portion of the graph just to get its
    #        topology
    GetVal(obj, VT)

    return set(children_iterator(obj, VT, child_VT, obj_type))


# -----------------------------------------------------------------------------
def leaves_iterator(obj_name, VT):
    stored = GetVal(obj_name, "StoredAttrs")

    for name, vt in onyx_dg.graph[(obj_name, VT)].children:
        if vt in stored:
            yield name, vt
        else:
            for leaf in leaves_iterator(name, vt):
                yield leaf


# -----------------------------------------------------------------------------
def LeafNodes(obj, VT):
    """
    Description:
        Return all leaf-level nodes (StoredAttrs) of a given node.
    Inputs:
        obj - object in database (or memory)
        VT  - value type
    Yields:
        A set of nodes.
    """
    if not isinstance(obj, str):
        obj = obj.Name

    # FIXME: have to evaluate this portion of the graph just to get its
    #        topology
    GetVal(obj, VT)

    return set(leaves_iterator(obj, VT))


# -----------------------------------------------------------------------------
def settable_iterator(obj_name, VT):
    for name, vt in onyx_dg.graph[(obj_name, VT)].children:
        node = onyx_dg.graph[(name, vt)]
        for leaf in settable_iterator(name, vt):
            if leaf is not None:
                yield leaf

        if isinstance(node, GraphNodeSettable):
            yield name, vt


# -----------------------------------------------------------------------------
def SettableNodes(obj, VT):
    """
    Description:
        Return all settable children of a given node (stopping at the first
        level).
    Inputs:
        obj - object in database (or memory)
        VT  - value type
    Yields:
        A set of nodes.
    """
    if not isinstance(obj, str):
        obj = obj.Name

    # FIXME: have to evaluate this portion of the graph just to get its
    #        topology
    GetVal(obj, VT)

    return set(settable_iterator(obj, VT))



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.database.objdb import ObjNotFound
from onyx.database.objdb_api import ObjDbQuery, GetObj
from onyx.database.ufo_base import BaseField, MutableField, FieldError
from onyx.datatypes.date import Date
from onyx.datatypes.curve import Curve, Knot
from onyx.datatypes.gcurve import GCurve, GKnot
from onyx.datatypes.hlocv import HlocvCurve, HlocvKnot
from onyx.datatypes.structure import Structure
from onyx.datatypes.table import Table
from onyx.datatypes.holiday_calendar import HolidayCalendar

import pickle
import base64
import gzip
import numbers

__all__ = [
    "StringField",
    "SelectField",
    "IntField",
    "SelectIntField",
    "FloatField",
    "BoolField",
    "BinaryField",
    "DateField",
    "ListField",
    "DictField",
    "SetField",
    "CurveField",
    "GCurveField",
    "HlocvCurveField",
    "StructureField",
    "HolidayCalField",
    "ReferenceField",
]


# -----------------------------------------------------------------------------
# --- serialize to plain string using base64 encoding
def pickle_loads(val):
    pickle.loads(gzip.decompress(base64.b64decode(val)))


def pickle_dumps(val):
    base64.b64encode(gzip.compress(pickle.dumps(val))).decode("utf-8")


###############################################################################
class StringField(BaseField):
    def validate(self, val):
        if not isinstance(val, str):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class SelectField(StringField):
    def __init__(self, *args, **kwds):
        self.options = kwds.pop("options", None)
        super().__init__(*args, **kwds)
        if not all([isinstance(opt, str) for opt in self.options]):
            types = [type(opt) for opt in self.options]
            raise FieldError("One or more optins are not "
                             "of type str: {0!s}".format(types))

    def validate(self, val):
        super().validate(val)
        if val not in self.options:
            raise FieldError("Illegal value {0!s}. Valid options "
                             "are {1!s}".format(val, self.options))


###############################################################################
class IntField(BaseField):
    def __init__(self, *args, **kwds):
        self.positive = kwds.pop("positive", False)
        super().__init__(*args, **kwds)

    def validate(self, val):
        if not isinstance(val, int):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        if self.positive and val < 0:
            raise FieldError("Input is not positive: {0:d}".format(val))


###############################################################################
class SelectIntField(IntField):
    def __init__(self, *args, **kwds):
        self.options = kwds.pop("options", None)
        super().__init__(*args, **kwds)
        if not all([isinstance(opt, int) for opt in self.options]):
            types = [type(opt) for opt in self.options]
            raise FieldError("One or more optins are not "
                             "of type int: {0!s}".format(types))

    def validate(self, val):
        super().validate(val)
        if val not in self.options:
            raise FieldError("Illegal value {0!s}. Valid options "
                             "are {1!s}".format(val, self.options))


###############################################################################
class FloatField(BaseField):
    def __init__(self, *args, **kwds):
        self.positive = kwds.pop("positive", False)
        super().__init__(*args, **kwds)

    def validate(self, val):
        if not isinstance(val, float):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        if self.positive and val < 0.0:
            raise FieldError("Input is not positive: {0:f}".format(val))


###############################################################################
class BoolField(BaseField):
    def validate(self, val):
        if not isinstance(val, bool):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class BinaryField(BaseField):
    def from_json(self, val):
        return gzip.decompress(base64.b64decode(val))

    def to_json(self, val):
        return base64.b64encode(gzip.compress(val)).decode("utf-8")

    def validate(self, val):
        if not isinstance(val, bytes):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class DateField(BaseField):
    def from_json(self, val):
        return Date.parse(val)

    def to_json(self, val):
        return val.isoformat()

    def validate(self, val):
        if not isinstance(val, Date):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class ListField(MutableField):
    def validate(self, val):
        if not isinstance(val, list):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class DictField(MutableField):
    def validate(self, val):
        if not isinstance(val, dict):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class SetField(MutableField):
    def from_json(self, val):
        return val if val is None else set(val)

    def to_json(self, val):
        return val if val is None else list(val)

    def validate(self, val):
        if not isinstance(val, set):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class CurveField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        crv = Curve.__new__(Curve)
        crv.knot_cls = Knot
        crv.data = [Knot(Date.parse(row["date"]), row["value"]) for row in val]
        return crv

    def to_json(self, crv):
        if crv is None:
            return crv
        else:
            return [{"date": knt.date.isoformat(),
                     "value": knt.value} for knt in crv]

    def validate(self, val):
        if not isinstance(val, Curve):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class GCurveField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        crv = GCurve.__new__(GCurve)
        crv.knot_cls = GKnot
        crv.data = [GKnot(Date.parse(d), pickle_loads(v)) for d, v in val]
        return crv

    def to_json(self, crv):
        if crv is None:
            return crv
        else:
            return [{"date": knt.date.isoformat(),
                     "value": pickle_dumps(knt.value)} for knt in crv]

    def validate(self, val):
        if not isinstance(val, GCurve):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class HlocvCurveField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        crv = HlocvCurve.__new__(HlocvCurve)
        crv.knot_cls = HlocvKnot
        crv.data = [HlocvKnot(Date.parse(d),
                              [float(v) for v in vls]) for d, vls in val]
        return crv

    def to_json(self, crv):
        if crv is None:
            return crv
        else:
            return [{"date": knt.date.isoformat(),
                     "value": knt.value} for knt in crv]

    def validate(self, val):
        if not isinstance(val, HlocvCurve):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class StructureField(MutableField):
    def from_json(self, val):
        return val if val is None else Structure(val)

    def to_json(self, val):
        return val if val is None else list(zip(val.keys(), val.values()))

    def validate(self, val):
        if not isinstance(val, Structure):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        # --- all keys should be strings
        if not all(isinstance(k, str) for k in val.keys()):
            FieldError("Structure keys need to be strings")
        # --- all values should be numbers
        if not all(isinstance(v, numbers.Number) for v in val.values()):
            FieldError("Structure values need to be numbers")


###############################################################################
class TableField(MutableField):
    def from_json(self, val):
        return val if val is None else pickle_loads(val)

    def to_json(self, val):
        return val if val is None else pickle_dumps(val)

    def validate(self, val):
        if not isinstance(val, Table):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class HolidayCalField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        else:
            return HolidayCalendar([Date.parse(d) for d in val])

    def to_json(self, val):
        return val if val is None else [d.isoformat() for d in val.holidays]

    def validate(self, val):
        if not isinstance(val, HolidayCalendar):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


# -----------------------------------------------------------------------------
def get_base_classes(cls):
    yield cls.__name__
    for base in cls.__bases__:
        yield from get_base_classes(base)


###############################################################################
class ReferenceField(StringField):
    def __init__(self, *args, **kwds):
        self.obj_type = kwds.pop("obj_type", None)
        super().__init__(*args, **kwds)

    def validate(self, val):
        if not isinstance(val, str):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        if self.obj_type is None:
            res = ObjDbQuery("""SELECT EXISTS "
                                (SELECT 1 FROM Objects
                                 WHERE ObjName=%s) AS "exists";""",
                             parms=(val,), attr="fetchone")
            if res.exists:
                raise FieldError("Object {0:s} "
                                 "not found in database".format(val))
        else:
            bases = set(get_base_classes(GetObj(val).__class__))

            if self.obj_type not in bases:
                msg = ("Object {0:s} exists in database, "
                       "but it is not instance of a subclass of {1:s}")
                raise FieldError(msg.format(val, self.obj_type))



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.database.ufo_base import UfoBase
from onyx.database.ufo_fields import IntField, FloatField, StringField
from onyx.depgraph.graph_api import GraphNodeVt

import time

__all__ = []


class TestCls0(UfoBase):
    def test_method(self, arg):
        return arg


class TestCls1(UfoBase):
    attr1 = IntField(default=333)
    attr2 = IntField(default=666)

    @GraphNodeVt("Property")
    def A(self, graph):
        return 1.0 + graph(self, "B") + graph(self, "C", 1, 2)

    @GraphNodeVt()
    def B(self, graph):
        # --- this is a slow method
        time.sleep(2.0)
        return graph(self, "attr1")

    @GraphNodeVt()
    def C(self, graph, x, y=0):
        # --- this is a slow calculated method
        time.sleep(2.0)
        return x + y + graph(self, "attr2")


class TestCls2(UfoBase):
    Kid1 = StringField()
    Kid2 = StringField()

    @GraphNodeVt()
    def A(self, graph):
        kid1 = graph(self, "Kid1")
        kid2 = graph(self, "Kid2")
        return graph(kid1, "A") + graph(kid2, "A")


class TestCls3(UfoBase):
    Number = FloatField(default=666.0)

    @GraphNodeVt("Property")
    def CalcProperty(self, graph):
        return 1.0 + 2.0*graph(self, "Number")



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.gcurve import GKnot, GCurve, KnotError, CurveError
from onyx.datatypes.curve import Curve

import numpy as np

__all__ = ["HlocvKnot", "HlocvCurve", "HlocvKnotError", "HlocvCurveError"]


# -----------------------------------------------------------------------------
class HlocvKnotError(KnotError):
    pass


# -----------------------------------------------------------------------------
class HlocvCurveError(CurveError):
    pass


###############################################################################
class HlocvKnot(GKnot):
    """
    Class representing a HLOCV knot. Used for daily stock prices. It inherits
    from GKnot.
    """
    __slots__ = ("date", "high", "low", "open", "close", "volume")

    # -------------------------------------------------------------------------
    def __init__(self, d, values):
        self.date = d
        self.high, self.low, self.open, self.close, self.volume = values

    # -------------------------------------------------------------------------
    def value_get(self):
        return (self.high, self.low, self.open, self.close, self.volume)

    def value_set(self, values):
        self.high, self.low, self.open, self.close, self.volume = values

    value = property(value_get, value_set)

    # -------------------------------------------------------------------------
    def __str__(self):
        fmt = "{0!s}, {1:12.5f}, {2:12.5f}, {3:12.5f}, {4:12.5f}, {5:12.0f}"
        return fmt.format(self.date, self.high,
                          self.low, self.open, self.close, self.volume)

    # -------------------------------------------------------------------------
    #  support serialization

    def __getstate__(self):
        return (self.date, self.high, self.low,
                self.open, self.close, self.volume)

    def __setstate__(self, state):
        (self.date, self.high,
         self.low, self.open, self.close, self.volume) = state


###############################################################################
class HlocvCurve(GCurve):
    """
    Class representing a HLOCV curve, generally used for daily stock prices. It
    inherits from the generic GCurve.
    """
    __slots__ = ("data", "knot_cls")

    # -------------------------------------------------------------------------
    def __init__(self, dates=None, values=None):
        super().__init__(dates, values, HlocvKnot)

    # -------------------------------------------------------------------------
    #  methods to return arrays for a specific HLOCV field:
    #    - highs
    #    - lows
    #    - opens
    #    - closes
    #    - volumes

    @property
    def highs(self):
        return np.array([knot.high for knot in self.data])

    @property
    def lows(self):
        return np.array([knot.low for knot in self.data])

    @property
    def opens(self):
        return np.array([knot.open for knot in self.data])

    @property
    def closes(self):
        return np.array([knot.close for knot in self.data])

    @property
    def volumes(self):
        return np.array([knot.volume for knot in self.data])

    # -------------------------------------------------------------------------
    # some methods inherited from GCurve don't make sense for an HLOCV curve

    def __add__(self, other):
        raise NotImplementedError()

    def __radd__(self, scalar):
        raise NotImplementedError()

    def __rmul__(self, scalar):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    def curve(self, field="close"):
        """
        Description:
            Convert a HLOCV curve to a scalar curve.
        Inputs:
            field - choose among high, low, open, close, volume
        Returns:
            A new Curve.
        """
        return Curve(self.dates, [getattr(knt, field) for knt in self.data])



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.date import Date
from onyx.datatypes.curve import Curve
from onyx.datatypes.gcurve import CurveError

import numpy as np
import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- create a first curve by passing arrays of dates and values
        self.dates = [Date.parse(d)
                      for d in ["F08", "G08", "H08", "J08", "K08", "M08",
                                "N08", "Q08", "U08", "V08", "X08", "Z08"]]
        self.values = np.cumsum(np.ones(12))
        self.refcrv = Curve(self.dates, self.values)

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_len(self):
        self.assertEqual(len(self.refcrv.data), 12)

    def test_repr(self):
        self.assertEqual(self.refcrv, eval(repr(self.refcrv)))

    def test_get(self):
        d0 = Date(2007, 6, 1)
        d1 = Date(2008, 6, 1)
        d2 = Date(2008, 6, 15)
        d3 = Date(2008, 7, 15)
        d4 = Date(2009, 6, 1)
        self.assertEqual(self.refcrv[d0], 1.0)              # use interpolation
        self.assertEqual(self.refcrv[d1], 6.0)              # knot exists
        self.assertEqual(self.refcrv[d2], 6.0 + 14.0/30.0)  # use interpolation
        self.assertEqual(self.refcrv[d3], 7.0 + 14.0/31.0)  # use interpolation
        self.assertEqual(self.refcrv[d4], 12.0)             # use interpolation
        # --- get a know that doesn't exists
        self.assertRaises(IndexError, self.refcrv.get_knot, d0)
        # --- get an existing knot
        self.assertEqual(self.refcrv.get_knot(d1).value, 6.0)
        # --- get a know that doesn't exists
        self.assertRaises(IndexError, self.refcrv.get_knot, d2)
        # --- get a know that doesn't exists
        self.assertRaises(IndexError, self.refcrv.get_knot, d3)
        # --- get a know that doesn't exists
        self.assertRaises(IndexError, self.refcrv.get_knot, d4)

    def test_set_and_del(self):
        value = 123
        date = Date.today()
        self.refcrv[date] = value
        self.assertEqual(self.refcrv.back.value, value)
        self.refcrv.del_knot(date)
        self.assertFalse(self.refcrv.has_knot(date))

    def test_has(self):
        self.assertFalse(self.refcrv.has_knot(Date(2007, 6, 1)))
        self.assertTrue(self.refcrv.has_knot(Date(2008, 6, 1)))
        self.assertFalse(self.refcrv.has_knot(Date(2008, 6, 15)))
        self.assertFalse(self.refcrv.has_knot(Date(2009, 6, 1)))

    def test_constructurs(self):
        # --- create a second curve by adding knots iteratively
        crv = Curve()
        val = 0
        for d in self.dates:
            val += 1
            crv[d] = val
        # --- do the comparison of their string representation
        self.assertEqual(crv.__str__(), self.refcrv.__str__())
        # --- do the comparison of their internal representation
        self.assertEqual(crv.data, self.refcrv.data)
        # --- do comparison between curves
        self.assertEqual(crv, self.refcrv)
        # --- check constructor exceptions
        d = Date.today()
        self.assertRaises(CurveError, Curve, [d, d], [1, 1])
        self.assertRaises(CurveError, Curve,
                          ([Date(2008, 1, 1), Date(2008, 12, 1)], [1, 1, 1]))

    def test_back_and_front(self):
        self.assertEqual(self.refcrv.front,
                         self.refcrv.knot_cls(Date(2008, 1, 1), 1.0))
        self.assertEqual(self.refcrv.front.date, Date(2008, 1, 1))
        self.assertEqual(self.refcrv.front.value, 1.0)
        self.assertEqual(self.refcrv.back,
                         self.refcrv.knot_cls(Date(2008, 12, 1), 12.0))
        self.assertEqual(self.refcrv.back.date, Date(2008, 12, 1))
        self.assertEqual(self.refcrv.back.value, 12.0)

    def test_dates_and_values(self):
        values = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        self.assertEqual(sum(self.refcrv.values - values), 0.0)
        dates = np.array([733042, 733073, 733102, 733133, 733163, 733194,
                          733224, 733255, 733286, 733316, 733347, 733377])
        self.assertEqual(sum(self.refcrv.ordinals - dates), 0.0)

    def test_algebra(self):
        values = self.refcrv.values
        self.assertEqual(2.0*self.refcrv - self.refcrv*1.0, self.refcrv)
        self.assertEqual(2.0*self.refcrv / 2.0, self.refcrv)
        self.assertEqual(sum(1.0 / values), sum((1.0 / self.refcrv).values))
        self.assertEqual(sum(values*values),
                         sum((self.refcrv*self.refcrv).values))

    def test_pickling(self):
        crv = pickle.loads(pickle.dumps(self.refcrv, 2))
        # --- check that data representations are identical
        self.assertEqual(crv.data, self.refcrv.data)
        # --- check that string representations are identical
        self.assertEqual(crv.__str__(), self.refcrv.__str__())

    def test_crop(self):
        sd = Date.parse("N08")
        ed = Date.parse("X08")
        dates = [Date.parse(d) for d in ["N08", "Q08", "U08", "V08", "X08"]]
        values = [7, 8, 9, 10, 11]
        crv = Curve(dates, values)
        self.assertEqual(self.refcrv.crop(sd, ed), crv)
        self.assertEqual(self.refcrv.crop(Date(2008, 6, 15),
                                          Date(2008, 11, 15)), crv)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.database.objdb_api import AddObj, DelObj, UseDatabase
from onyx.depgraph.graph import (DependencyGraph, CreateNode, GraphNode,
                                 GraphNodeSettable, GraphNodeCalc)
from onyx.depgraph.test import ufo_testcls

import onyx.depgraph as onyx_dg

import unittest
import copy
import types
import os


class RegTest(unittest.TestCase):
    def setUp(self):
        self.use_db = UseDatabase(os.getenv("OBJDB_TEST", default="TestDb"))
        self.use_db.__enter__()
        onyx_dg.graph = DependencyGraph()
        self.test_obj = AddObj(ufo_testcls.TestCls0(Name="testobj"))

    def tearDown(self):
        DelObj(self.test_obj)
        del self.test_obj
        self.use_db.__exit__()

    def test_create_nodes(self):
        # --- create a node that points to "Name"
        node = CreateNode((self.test_obj.Name, "Name"))
        self.assertTrue(isinstance(node, GraphNodeSettable))
        self.assertEqual(node.get_value(), "testobj")

        # --- create a node that points to "Version"
        node = CreateNode((self.test_obj.Name, "Version"))
        self.assertTrue(isinstance(node, GraphNodeSettable))
        self.assertEqual(node.get_value(), 0)

        # --- create a node that points to "StoredAttrs"
        node = CreateNode((self.test_obj.Name, "StoredAttrs"))
        self.assertTrue(isinstance(node, GraphNode))
        self.assertEqual(node.get_value(), self.test_obj.StoredAttrs)

        # --- create a node that points to "test_method"
        node = CreateNode((self.test_obj.Name, "test_method"))
        self.assertTrue(isinstance(node, GraphNodeCalc))
        self.assertEqual(node.get_value(123), self.test_obj.test_method(123))

        # --- create a node that points to a non-existing method
        self.assertRaises(AttributeError, CreateNode,
                          (self.test_obj.Name, "not_there"))

    def test_invalidation1(self):
        # --- create a node that points to "StoredAttrs"
        node = CreateNode((self.test_obj.Name, "StoredAttrs"))

        test_obj_attrs = copy.deepcopy(self.test_obj.StoredAttrs)

        # --- as before, test_obj has only standard attributes
        self.assertEqual(node.get_value(), test_obj_attrs)

        # --- add a new attribute to test_obj
        self.test_obj.NewAttr = "This is brand new!"

        # --- the node is still valid, hence get_value still returns the same
        #     set of attributes as before
        self.assertTrue(node.valid)
        self.assertEqual(node.get_value(), test_obj_attrs)

        # --- after invalidation we expect to see the new attribute
        node.invalidate()
        self.assertFalse(node.valid)
        self.assertEqual(node.get_value(), self.test_obj.StoredAttrs)
        self.assertTrue(node.valid)

    def test_invalidation2(self):
        # --- we add a few more attributes to test_obj as well as a
        #     calculated method
        self.test_obj.Num1 = 1.0
        self.test_obj.Num2 = 2.0

        setattr(ufo_testcls.TestCls0, "sum",
                types.MethodType(lambda s: s.Num1 + s.Num2, self.test_obj))

        # --- create a node that points to "sum"
        node = CreateNode((self.test_obj.Name, "sum"))

        # --- this should be a calculated node whose value is Num1+Num2
        self.assertTrue(isinstance(node, GraphNodeCalc))
        self.assertEqual(node.get_value(), 3.0)

        # --- if we change Num1, nothing should happen to sum because the
        #     result is still cached
        self.test_obj.Num1 = 2.0
        self.assertTrue(node.valid)
        self.assertEqual(node.get_value(), 3.0)

        # --- but after invalidation, sum will be recalculated
        node.invalidate()
        self.assertFalse(node.valid)
        self.assertEqual(node.get_value(), 4.0)

    def test_clone(self):
        # --- create a node that points to "StoredAttrs" and make a clone
        node1 = CreateNode((self.test_obj.Name, "StoredAttrs"))
        node2 = node1.clone()

        # --- test that the two nodes are different objects, return the same
        #     value and share the same reference to the underlying object
        self.assertNotEqual(node1, node2)
        self.assertEqual(node1.get_value(), node2.get_value())
        self.assertEqual(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref(), self.test_obj)
        self.assertIs(node2.obj_ref(), self.test_obj)

        ## --- again, we add a few more attributes to test_obj as well as a
        #     calculated method
        self.test_obj.Num1 = 1.0
        self.test_obj.Num2 = 2.0

        def my_sum(instance):
            return instance.Num1 + instance.Num2

        setattr(ufo_testcls.TestCls0, "sum",
                types.MethodType(my_sum, self.test_obj))

        # --- create a node that points to "sum" and make a clone
        node1 = CreateNode((self.test_obj.Name, "sum"))
        node2 = node1.clone()

        # --- test that the two nodes are different objects, return the same
        #     value and share the same reference to the underlying object
        self.assertNotEqual(node1, node2)
        self.assertEqual(node1.get_value(), node2.get_value())
        self.assertEqual(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref(), self.test_obj)
        self.assertIs(node2.obj_ref(), self.test_obj)

        # --- test that invalidation only affects a specific node
        self.test_obj.Num1 = 2.0
        node1.invalidate()
        self.assertFalse(node1.valid)
        self.assertTrue(node2.valid)
        self.assertEqual(node1.get_value(), 4.0)
        self.assertEqual(node2.get_value(), 3.0)
        node2.invalidate()
        self.assertFalse(node2.valid)
        self.assertEqual(node2.get_value(), 4.0)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.table import Table

import numpy as np
import itertools
import string
import io

__all__ = [
    "TableExtract",
    "TableFilter",
    "TableFuncApply",
    "TableFuncApplyToColumn",
    "TableGroupBy",
    "TableFromCurves",
    "TableFromCsv",
    "TableToCsv",
]


# -----------------------------------------------------------------------------
def TableExtract(table, columns, new_names=None):
    """
    Description:
        Return a new sub-Table with the required columns.
    Inputs:
        table     - the input table
        columns   - a list of names of the columns to be extracted
        new_names - a list of new column names
    Returns:
        A new Table.
    """
    new_names = new_names or columns
    new_table = [new_names] + \
                [[row.__getattribute__(k) for k in columns] for row in table]

    return Table(new_table)


# -----------------------------------------------------------------------------
def TableFilter(table, func, in_place=True):
    """
    Description:
        Remove rows from the table based on a user-defined filter function
    Inputs:
        table    - the input table
        func     - user-defined function. Takes a row as input and returns
                   False if the row has to be excluded.
        in_place - if False, create a new table
    Returns:
        A new Table if in_place is False
    """
    if in_place:
        table.data = [row for row in table.data if func(row)]
        return table
    else:
        return Table([list(table.tags)] +
                     [row for row in table.data if func(row)])


# -----------------------------------------------------------------------------
def TableFuncApply(table, func, col_name):
    """
    Description:
        Return a new table with an extra column whose values are obtained
        applying the user-defined function to each row of the table.
    Inputs:
        table    - the input table
        func     - the user-defined function
        col_name - the name of the new column
    Returns:
        A new Table.
    """
    new_table = [list(table.tags) + [col_name]] + \
                [row.values + [func(row)] for row in table]

    return Table(new_table)


# -----------------------------------------------------------------------------
def TableFuncApplyToColumn(table, func, col_name):
    """
    Description:
        Applie in-place a user-defined function to a column of the table.
    Inputs:
        table    - the input table
        func     - the user-defined function
        col_name - the name of the column that is modified by the user-defined
                   function
    """
    for row in table.data:
        row[col_name] = func(row[col_name])
    return table


# -----------------------------------------------------------------------------
def TableGroupBy(table, pivot, measures):
    """
    Description:
        Create a pivot-table from the input table.
    Inputs:
        table    - the input table
        pivot    - column name used for vertical pivot aggregation
        measures - list of tuples in the form ( key, funcname ) where:
                     - key is the name of the column to be used as measure
                     - funcname is the aggregation function (SUM, COUNT,
                       AVG, STD, MIN, MAX)
    Returns:
        A new table.
    """
    # --- sort the table by vertical pivot
    table.sort((pivot, ))

    # --- key-function for vertical pivot
    keyfunc = lambda r: r[pivot]

    # --- arrays of tags and measure functions for pivot table
    tags = [pivot]
    fns = []
    for mkey, funcname in measures:
        funcname = funcname.upper()
        tags.append("{0:s}_{1:s}".format(mkey, funcname))
        if funcname == "SUM":
            fns.append(lambda g: sum([r[mkey] for r in g]))
        elif funcname == "COUNT":
            fns.append(lambda g: len([r[mkey] for r in g]))
        elif funcname == "AVG":
            fns.append(lambda g: np.mean([r[mkey] for r in g]))
        elif funcname == "STD":
            fns.append(lambda g: np.std([r[mkey] for r in g]))
        elif funcname == "MIN":
            fns.append(lambda g: min([r[mkey] for r in g]))
        elif funcname == "MAX":
            fns.append(lambda g: max([r[mkey] for r in g]))
        else:
            raise NameError("Unrecognized grouping "
                            "function {0:s}".format(funcname))

    # --- put together the pivot table
    pivot_table = [tags]
    for key, group in itertools.groupby(table.data, keyfunc):
        group = list(group)
        pivot_table.append([key] + [f(group) for f in fns])

    return Table(pivot_table)


# -----------------------------------------------------------------------------
def TableFromCurves(crvs, tags=None):
    """
    Description:
        Convert a list of curves into a table.
    Inputs:
        crvs - a list of curves
        tags - a list of names (Optional)
    Returns:
        A Table.
    """
    dates = set()
    for crv in crvs:
        dates.update(crv.dates)

    if tags is None:
        tab = [["Date"] + ["Curve{0:02d}".format(i) for i in range(len(crvs))]]
    else:
        tab = [["Date"] + tags]

    for d in sorted(dates):
        values = [d]
        for crv in crvs:
            try:
                knot = crv.get_knot(d)
                try:
                    values.append(knot.close)
                except AttributeError:
                    values.append(knot.value)
            except IndexError:
                values.append(None)
        tab.append(values)

    return Table(tab)


# -----------------------------------------------------------------------------
def TableFromCsv(file_name="", delimiter=","):
    """
    Description:
        Create a table from a csv text file. The first line of the file must
        define all the tags for the table (header).
    Inputs:
        file_name - the name of the csv file. The full path must be specified.
        delimiter - the character used to separate fields
    Returns:
        A Table.
    """
    tbl = []
    with open(file_name, "r") as csvfile:
        for line in csvfile:
            # --- split line, eliminating EOL and quotes and add to the
            #     table array
            row = line.rstrip().replace("\"", "").split(delimiter)
            tbl.append([r.strip() for r in row])

    # --- remove spaces and illegal characters form header
    trantab = string.maketrans("", "")
    header = tbl[0]
    for i, item in enumerate(header):
        header[i] = str(item).translate(trantab, " ?%/.=:")
    tbl[0] = header

    return Table(tbl)


# -----------------------------------------------------------------------------
def TableToCsv(table, file_name=None, delimiter=",", use_quotes=False):
    """
    Description:
        Write the content of a Table to a text file in csv format.
    Inputs:
        table      - the input table
        file_name  - the name of the HTML file. The full filepath must be
                     specified. If None, cStringIO.StringIO() is used and a
                     string with the csv table is returned.
        delimiter  - the character used to separate fields
        use_quotes - if True, wrap every record in double quotes
    Returns:
        A string with the Table in csv format if file_name is None.
    """
    if use_quotes:
        fstr = lambda x: '"{0:s}"'.format(str(x).encode("cp850", "replace"))
    else:
        fstr = lambda x: "{0:s}".format(str(x).encode("cp850", "replace"))

    if file_name is None:
        csvfile = io.StringIO()
    else:
        csvfile = open(file_name, "w")

    try:
        csvfile.writelines(delimiter.join(table.tags) + "\n")
        for row in table.data:
            csvfile.writelines(delimiter.join([fstr(item)
                                               for item in row.values]) + "\n")
        # --- if writing to StringIO, copy content into a string before
        #     releasing the resource
        if file_name is None:
            output = csvfile.getvalue()
        else:
            output = None
    finally:
        csvfile.close()

    return output



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################


from onyx.database.objdb import ObjNotFound
from onyx.depgraph.graph import GraphNodeError
from onyx.depgraph.graph_api import (CreateInMemory,
                                     GetVal, SetVal, InvalidateNode, UseGraph)
from onyx.database.objdb_api import UseDatabase
from onyx.depgraph.test import ufo_testcls

import time
import unittest
import os


class timer(object):
    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.elapsed = self.end - self.start
        return False


# --- Unit tests class
class TestGraph(unittest.TestCase):
    def setUp(self):
        self.use_db = UseDatabase(os.getenv("OBJDB_TEST", default="TestDb"))
        self.use_graph = UseGraph()
        self.use_db.__enter__()
        self.use_graph.__enter__()

        self.obj1 = CreateInMemory(ufo_testcls.TestCls1(Name="obj1"))
        self.obj2 = CreateInMemory(ufo_testcls.TestCls1(Name="obj2"))
        self.obj3 = CreateInMemory(ufo_testcls.TestCls2(Name="obj3",
                                                        Kid1=self.obj1.Name,
                                                        Kid2=self.obj2.Name))

    def tearDown(self):
        del self.obj1, self.obj2, self.obj3
        self.use_db.__exit__()
        self.use_graph.__exit__()

    def test_GetVal(self):
        # --- check that we retrive the correct values
        self.assertEqual(GetVal("obj1", "attr1"), 333.0)
        self.assertEqual(GetVal("obj1", "attr2"), 666.0)
        self.assertEqual(GetVal("obj1", "A"), 1003.0)
        self.assertEqual(GetVal(self.obj1, "A"), 1003.0)
        self.assertEqual(GetVal("obj1", "C", 0, 0), 666.0)
        self.assertEqual(GetVal("obj1", "C", 1, 0), 667.0)
        self.assertEqual(GetVal("obj1", "C", x=1), 667.0)
        self.assertEqual(GetVal("obj1", "C", 0, 2), 668.0)
        self.assertEqual(GetVal("obj1", "C", 0, y=2), 668.0)
        self.assertEqual(GetVal("obj3", "A"), 2006.0)
        self.assertEqual(GetVal("obj1", "A"), GetVal("obj1", "A"))

    def test_caching(self):
        # --- compute A the first time
        with timer() as t:
            val = GetVal("obj1", "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 4.0, 1)

        # --- compute A a second time: result is cached
        with timer() as t:
            val = GetVal("obj1", "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 0.0, 1)

        # --- invalidate C and recalculate A: should only have to recalculate
        #     one of the two children
        InvalidateNode(self.obj1, "C")
        with timer() as t:
            val = GetVal("obj1", "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 2.0, 1)

        # --- try once more: should be cached again
        with timer() as t:
            val = GetVal("obj1", "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 0.0, 1)

    def test_exceptions(self):
        self.assertRaises(ObjNotFound, GetVal, "xxx", "xxx")
        self.assertRaises(AttributeError, GetVal, "obj1", "xxx")
        self.assertRaises(GraphNodeError, SetVal, "obj1", "A", "Hello World")
        self.assertRaises(GraphNodeError, SetVal, "obj1", "B", 0)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.date import Date
from onyx.database.ufo_base import SPECIAL

import psycopg2 as psql
import psycopg2.extras as psql_extras
import psycopg2.extensions as psql_ext
import psycopg2.errorcodes as psql_err
import pylru
import weakref
import pickle
import getpass
import abc

__all__ = ["ObjNotFound", "ObjDbError", "ObjExists", "ObjDbClient"]

USER = getpass.getuser()

# --- define standard query strings and error messages

QUERY_CLS_INSERT = "INSERT INTO ClassDefinitions VALUES (%s,%s);"
QUERY_CLS_GET = """
SELECT ClassDef FROM ClassDefinitions WHERE ObjType=%s;"""
QUERY_CLS_DELETE = """
DELETE FROM ClassDefinitions row WHERE row.ObjType=%s AND NOT
EXISTS (SELECT 1 FROM Objects WHERE ObjType=row.ObjType);"""

QUERY_OBJ_INSERT = "INSERT INTO Objects VALUES (%s,%s,%s,%s,%s,%s,%s);"
QUERY_OBJ_GET = """
SELECT Name, ObjType, Version, ChangedBy, TimeCreated, LastUpdated, Data
FROM Objects WHERE Name=%s;"""
QUERY_OBJ_UPDATE = """
UPDATE Objects
SET Version=%s, ChangedBy=%s, LastUpdated=%s, Data=%s
WHERE Name=%s AND Version=%s;"""
QUERY_OBJ_DELETE = """
WITH deleted_row AS (
    DELETE FROM Objects
    WHERE  Name=%s AND Version=%s
    RETURNING ObjType, Name)
DELETE FROM ClassDefinitions cls_def
USING deleted_row
WHERE cls_def.ObjType = deleted_row.ObjType
AND NOT EXISTS (
   SELECT 1 FROM Objects objs
   WHERE objs.ObjType = deleted_row.ObjType
   AND objs.Name <> deleted_row.Name );"""
QUERY_OBJ_VERSION = "SELECT Version FROM Objects WHERE Name=%s LIMIT 1;"
QUERY_OBJ_EXISTS = """
SELECT EXISTS (SELECT 1 FROM Objects WHERE Name=%s) AS "exists";"""

QUERY_ARC_INSERT = "INSERT INTO Archive VALUES (%s,%s,%s,%s,%s,%s);"

MSG_RELOAD = """
Instance of {0:s} is older (v.{1:d}) than the most recent version in database
(v.{2:d}): reload object from database before updating"""


###############################################################################
class ObjNotFound(Exception):
    pass


###############################################################################
class ObjDbError(Exception):
    pass


###############################################################################
class ObjExists(ObjDbError):
    pass


###############################################################################
class dummy_conn(object):
    """
    This is a dummy connection class that implements placeholders methods
    used to support transactions. It is used by the base and dummy clients
    and should not be used by production implementations.
    """
    def set_isolation_level(self, *args):
        pass

    def tpc_begin(self, *args):
        pass

    def tpc_commit(self, *args):
        pass

    def tpc_rollback(self, *args):
        pass


###############################################################################
class ObjDbBase(object):
    """
    Base class for ObjDb clients.
    """
    # -------------------------------------------------------------------------
    def __init__(self, database, user, host=None):
        self.dbname = database
        self.user = user
        self.host = host

        # --- weak references to all queried objects are kept here. Subsequent
        #     queries for the same object will return a reference if the object
        #     is still alive (i.e. they have not been disposed of by the gc).
        #     weak references will be kept alive by:
        #       - the cache of the db-client
        #       - the cache of the dependency graph
        self.refs = weakref.WeakValueDictionary()

        # --- unlimited caching
        self.cache = {}

        # --- lookup dictionary for class definitions
        self.class_defs = {}

        # --- the connection to the db-backend
        self.conn = dummy_conn()

        # --- store current transaction name
        self.transaction = None

    # -------------------------------------------------------------------------
    def class_lookup(self, cls_name):
        """
        Lookup class by class name.
        """
        return self.class_defs[cls_name]

    # -------------------------------------------------------------------------
    #  implements methods to retrieve, add and remove items from db-cache

    def __getitem__(self, name):
        return self.refs[name]

    def __setitem__(self, name, instance):
        self.cache[name] = instance
        self.refs[name] = instance

    def __delitem__(self, name):
        try:
            del self.cache[name]
            del self.refs[name]
        except KeyError:
            pass

    # -------------------------------------------------------------------------
    #  method to test if an object exists in database. the base class
    #  implementation only checks if a reference exsists without testing the
    #  backend.
    def __contains__(self, name):
        return name in self.refs

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def initialize(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def restart(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def close(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def add(self, obj):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def get(self, name, refresh=False):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def update(self, obj):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def delete(self, obj):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def add_dated(self, obj, date):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def get_dated(self, name, date, strict=False):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def update_dated(self, obj, date, override=False):
        raise NotImplementedError()


# -----------------------------------------------------------------------------
def typecast_to_Date(value, curs):
    if value is None:
        return None
    else:
        # --- maybe we should use a regexp like r"^.*'(.*)'.*$" ???
        return Date.parse(value)


###############################################################################
class ObjDbClient(ObjDbBase):
    core_tables = ("ClassDefinitions", "Objects", "Archive")

    # -------------------------------------------------------------------------
    def __init__(self, database, user, host=None, cache_size=500, check=True):
        """
        Description:
            ObjDb database, client class. It exposes a minimal api.
        """
        super().__init__(database, user, host)

        # --- The database cache is used to keep a limited number of previously
        #     retrieved objects "alive" (see above).
        self.cache = pylru.lrucache(cache_size)

        # --- open database connection
        self.conn = psql.connect(host=host, database=database, user=user,
                                 cursor_factory=psql_extras.NamedTupleCursor)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

        # --- register typecast to Date for date, timpestamp, timestamptz
        curs = self.conn.cursor()
        curs.execute("SELECT NULL::date, NULL::timestamp, NULL::timestamptz;")
        oids = tuple([col.type_code for col in curs.description])
        psql_ext.register_type(psql_ext.new_type(oids, "Date",
                                                 typecast_to_Date))

        # --- register a no-op loads() function to turn-off automatic parsing
        #     of json/jsonb datatype
        psql_extras.register_default_json(loads=lambda x: x)
        psql_extras.register_default_jsonb(loads=lambda x: x)

        # --- validation: without the following tables in the backend the
        #                 client doesn't work properly.
        if check:
            curs = self.conn.cursor()
            for table in self.core_tables:
                try:
                    curs.execute("SELECT 1 FROM {0:s} LIMIT 1;".format(table))
                except psql.ProgrammingError:
                    raise RuntimeError("Table {0!s} is missing on "
                                       "{1!s}".format(table, self.dbname))

    # -------------------------------------------------------------------------
    def initialize(self):
        curs = self.conn.cursor()
        curs.execute("""
            -- this table is used to lookup class definition from class name
            CREATE TABLE
            ClassDefinitions (
                ObjType  varchar(64) PRIMARY KEY,
                ClassDef bytea       NOT NULL );

            -- this table is used to store UFO objects
            CREATE TABLE
            Objects (
                Name        varchar(64) PRIMARY KEY,
                ObjType     varchar(64) REFERENCES ClassDefinitions,
                Version     integer     NOT NULL,
                ChangedBy   varchar(64) NOT NULL,
                TimeCreated timestamptz NOT NULL,
                LastUpdated timestamptz NOT NULL,
                Data        jsonb       NOT NULL );

            CREATE INDEX objects_objtype_idx ON Objects (ObjType);

            -- this table is used to store archived (i.e. dated) objects
            CREATE TABLE
            Archive (
                Name      varchar(64) REFERENCES Objects,
                ObjType   varchar(64) REFERENCES ClassDefinitions,
                Date      date        NOT NULL,
                ChangedBy varchar(64) NOT NULL,
                TimeStamp timestamptz NOT NULL,
                Data      jsonb       NOT NULL,
                CONSTRAINT name_date UNIQUE (Name, Date) );

            CREATE INDEX archive_name_idx ON Archive (Name);""")

    # -------------------------------------------------------------------------
    def class_lookup(self, cls_name):
        try:
            return self.class_defs[cls_name]
        except KeyError:
            curs = self.conn.cursor()
            curs.execute(QUERY_CLS_GET, (cls_name,))
            row = curs.fetchone()
            if row is None:
                raise KeyError("Class definition unavailable for"
                               "{0!s} in {1!s}".format(cls_name, self.dbname))
            else:
                cls = pickle.loads(bytes(row[0]))
                self.class_defs[cls_name] = cls
                return cls

    # -------------------------------------------------------------------------
    def __contains__(self, name):
        if name in self.refs:
            return True
        curs = self.conn.cursor()
        curs.execute(QUERY_OBJ_EXISTS, (name,))
        return curs.fetchone().exists

    # -------------------------------------------------------------------------
    def restart(self):
        self.close()
        self.conn = psql.connect(host=self.host,
                                 database=self.dbname, user=self.user)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

    # -------------------------------------------------------------------------
    def close(self):
        if hasattr(self, "conn") and not self.conn.closed:
            self.conn.close()

    # -------------------------------------------------------------------------
    def add(self, obj):
        curs = self.conn.cursor()

        # --- validate object's Name
        if not isinstance(obj.Name, str):
            if obj.Name is None:
                raise ObjDbError("Cannot persist objects with undefined Name")
            else:
                raise ObjDbError("Name attribute must be "
                                 "a string: {0!s}".format(type(obj.Name)))

        # --- if class definition is missing, add it to ClassDefinitions table
        try:
            cls = self.class_lookup(obj.ObjType)
        except KeyError:
            cls = psql.Binary(pickle.dumps(obj.__class__,
                                           protocol=pickle.HIGHEST_PROTOCOL))
            curs.execute(QUERY_CLS_INSERT, (obj.ObjType, cls))

        # --- serialize object and create record
        record = (obj.Name, obj.ObjType, obj.Version, obj.ChangedBy,
                  obj.TimeCreated, obj.LastUpdated, obj.to_json())

        # --- DB Insert
        try:
            curs.execute(QUERY_OBJ_INSERT, record)
        except psql.IntegrityError as err:
            if psql_err.lookup(err.pgcode) == "UNIQUE_VIOLATION":
                raise ObjExists("Object {0!s} already in "
                                "{1!s}".format(obj.Name, self.dbname))
            else:
                raise ObjDbError("{0!s}. Name = "
                                 "{1!s}".format(err.message, obj.Name))
        finally:
            curs.close()

        self[obj.Name] = obj
        return obj

    # -------------------------------------------------------------------------
    def get(self, name, refresh=False):
        try:
            # --- if object must be retrieved from database raise exception and
            #     fall back
            if refresh:
                raise KeyError()
            # --- then try getting obj from database cache
            return self[name]
        except KeyError:
            # --- if not found in the db cache, get it from the database
            curs = self.conn.cursor()
            curs.execute(QUERY_OBJ_GET, (name,))
            row = curs.fetchone()
            curs.close()

            if row is None:
                raise ObjNotFound("Object '{0!s}' not found "
                                  "in {1!s}".format(name, self.dbname))
            else:
                cls = self.class_lookup(row.objtype)
                obj = cls.__new__(cls)

                for attr in SPECIAL:
                    setattr(obj, attr, getattr(row, attr.lower()))

                obj.from_json(row.data)

            self[obj.Name] = obj
            return obj

    # -------------------------------------------------------------------------
    def update(self, obj):
        curs = self.conn.cursor()

        # --- current version is used by the update query
        curr_ver = obj.Version

        # --- update time-stamp and version on the object
        obj.LastUpdated = Date.now()
        obj.ChangedBy = USER.upper()
        obj.Version += 1

        # --- serialize data and create record
        record = (obj.Version, obj.ChangedBy,
                  obj.LastUpdated, obj.to_json(), obj.Name, curr_ver)

        try:
            curs.execute(QUERY_OBJ_UPDATE, record)

            if curs.rowcount == 0:
                # --- upadte failed: check why
                curs.execute(QUERY_OBJ_VERSION, (obj.Name,))
                res = curs.fetchone()

                if res is None:
                    raise ObjNotFound("Object '{0!s}' not found "
                                      "in {1!s}".format(obj.Name, self.dbname))

                if curr_ver < res.version:
                    msg = MSG_RELOAD.format(obj.Name, curr_ver, res.version)
                    raise ObjDbError(msg)

            else:
                self[obj.Name] = obj
                return obj

        except psql.IntegrityError as err:
            raise ObjDbError("{0!s}. Name = "
                             "{1!s}".format(err.message, obj.Name))
        finally:
            curs.close()

    # -------------------------------------------------------------------------
    def delete(self, obj):
        # --- call the delete method if available
        if hasattr(obj, "delete"):
            obj.delete()

        try:
            # --- DB Delete
            curs = self.conn.cursor()
            curs.execute(QUERY_OBJ_DELETE, (obj.Name, obj.Version))

            if curs.rowcount == 0:
                # --- delete failed: check why
                curs.execute(QUERY_OBJ_VERSION, (obj.Name,))
                res = curs.fetchone()

                if res is None:
                    raise ObjNotFound("Object '{0!s}' not found "
                                      "in {1!s}".format(obj.Name, self.dbname))
                elif obj.Version < res.version:
                    msg = MSG_RELOAD.format(obj.Name, obj.Version, res.version)
                    raise ObjDbError(msg)
                else:
                    raise ObjDbError("Deletion of {0!s} failed without "
                                     "an obvious cause...".format(obj.Name))
        finally:
            curs.close()
            del self[obj.Name]

    # -------------------------------------------------------------------------
    def add_dated(self, obj, date):
        curs = self.conn.cursor()

        # --- validate object's Name
        if not isinstance(obj.Name, str):
            if obj.Name is None:
                raise ObjDbError("Cannot persist objects with undefined Name")
            else:
                raise ObjDbError("Name attribute must be "
                                 "a string: {0!s}".format(type(obj.Name)))

        # --- if class definition is missing, add it to ClassDefinitions table
        try:
            cls = self.class_lookup(obj.ObjType)
        except KeyError:
            cls = psql.Binary(pickle.dumps(obj.__class__,
                                           protocol=pickle.HIGHEST_PROTOCOL))
            curs.execute(QUERY_CLS_INSERT, (obj.ObjType, cls))

        # --- serialize object and create record
        record = (obj.Name, obj.ObjType, date,
                  obj.ChangedBy, obj.LastUpdated, obj.to_json())

        # --- DB Insert
        try:
            curs.execute(QUERY_ARC_INSERT, record)
        except psql.IntegrityError as err:
            if psql_err.lookup(err.pgcode) == "UNIQUE_VIOLATION":
                raise ObjExists("Object {0!s} already in "
                                "{1!s}".format(obj.Name, self.dbname))
            else:
                raise ObjDbError("{0!s}. Name = "
                                 "{1!s}".format(err.message, obj.Name))
        finally:
            curs.close()

        self[(obj.Name, date)] = obj
        return obj

    # -------------------------------------------------------------------------
    def get_dated(self, name, date, strict=False):
        try:
            # --- then try getting obj from database cache
            return self[(name, date)]
        except KeyError:
            if strict:
                query = """SELECT
                           ObjType, Date, ChangedBy, TimeStamp, Data
                           FROM Archive
                           WHERE Name = %s AND Date = %s;"""
                parms = (name, date)
            else:
                query = """SELECT
                           ObjType, Date, ChangedBy, TimeStamp, Data
                           FROM Archive
                           WHERE Name = %s AND Date IN
                               (SELECT MAX(Date) FROM Archive
                                WHERE Name = %s AND Date <= %s);"""
                parms = (name, name, date)

            curs = self.conn.cursor()
            curs.execute(query, parms)
            row = curs.fetchone()
            curs.close()

            if row is None:
                raise ObjNotFound("Archived object '{0!s}' not "
                                  "found for date {1!s}".format(name, date))
            else:
                cls = self.class_lookup(row.objtype)
                obj = cls.__new__(cls)

                setattr(obj, "Name", name)
                setattr(obj, "ObjType", row.objtype)
                setattr(obj, "ChangedBy", row.changedby)
                setattr(obj, "LastUpdated", row.timestamp)

                obj.from_json(row.data)

            if not strict:
                # --- we always  want to cache the object based on its actual
                #     archive date
                date = Date.parse(row.date)

            self[(name, date)] = obj
            return obj

    # -------------------------------------------------------------------------
    def update_dated(self, obj, date, overwrite=False):
        curs = self.conn.cursor()

        # --- check whether the date in the object matches the timestamp
        if date != obj.LastUpdated:
            ObjDbError("Inconsistent dates for {0!s}: date is {1!s}, "
                       "LastUpdated is {2!s}".format(obj.Name, date,
                                                     obj.LastUpdated))

        # --- check whether the dated object exists
        query = """SELECT Name FROM Archive
                   WHERE Name=%s AND Date=%s LIMIT 1;"""
        curs.execute(query, (obj.Name, date))
        if curs.fetchone() is None:
            raise ObjNotFound("Archived object '{0!s}' not found in {1!s} "
                              "for date {2!s}".format(obj.Name,
                                                      self.dbname, date))

        if not overwrite:
            # --- check if trying to update an object archived in the past
            if date < Date.today():
                raise ObjDbError("Trying to update an "
                                 "archived object without permissions")

        # --- update time-stamp and version on the object
        obj.LastUpdated = Date.now()
        obj.ChangedBy = USER.upper()

        # --- serialize object and create record
        record = (obj.ChangedBy,
                  obj.LastUpdated, obj.to_json(), obj.Name, date)

        # --- DB Insert
        try:
            curs.execute("""UPDATE Archive
                            SET ChangedBy = %s, TimeStamp = %s, Data = %s
                            WHERE Name = %s AND Date = %s;""", record)
        except psql.IntegrityError as err:
            if psql_err.lookup(err.pgcode) == "UNIQUE_VIOLATION":
                raise ObjExists("Dated Object {0!s}-{1!s} already in "
                                "{2!s}".format(obj.Name, date, self.dbname))
            else:
                raise ObjDbError("{0!s}. Name = {1!s}, date = "
                                 "{2!s}".format(err.message, obj.Name, date))
        finally:
            curs.close()

        self[(obj.Name, date)] = obj
        return obj


###############################################################################
class ObjDbDummyClient(ObjDbBase):
    """
    A dummy client that never queries the database backend.
    """
    # -------------------------------------------------------------------------
    def restart(self):
        pass

    # -------------------------------------------------------------------------
    def close(self):
        pass

    # -------------------------------------------------------------------------
    def add(self, obj):
        if obj.Name in self:
            raise ObjExists("Object {0!s} already "
                            "in {1!s}".format(obj.Name, self.dbname))
        self[obj.Name] = obj
        return obj

    # -------------------------------------------------------------------------
    def get(self, name, refresh=False):
        if refresh:
            raise ObjDbError("get method of "
                             "ObjDbDummyClient does not support refresh")
        try:
            return self[name]
        except KeyError:
            raise ObjNotFound("Object '{0!s}' "
                              "not found in {1!s}".format(name, self.dbname))

    # -------------------------------------------------------------------------
    def update(self, obj):
        try:
            self[obj.Name] = obj
        except KeyError:
            raise ObjNotFound("Object '{0!s}' not "
                              "found in {1!s}".format(obj.Name, self.dbname))
        return obj

    # -------------------------------------------------------------------------
    def delete(self, obj):
        if hasattr(obj, "delete"):
            obj.delete()
        del self[obj.Name]

    # -------------------------------------------------------------------------
    def add_dated(self, obj, date):
        self[(obj.Name, date)] = obj
        return obj

    # -------------------------------------------------------------------------
    def get_dated(self, name, date, strict=False):
        try:
            return self[(name, date)]
        except KeyError:
            raise ObjNotFound("Archived object '{0!s}' not "
                              "found for date {1!s}".format(name, date))

    # -------------------------------------------------------------------------
    def update_dated(self, obj, date, override=False):
        self[(obj.Name, date)] = obj
        return obj



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.date import Date
from onyx.datatypes.holiday_calendar import HolidayCalendar

from dateutil.relativedelta import relativedelta, MO, FR

import datetime
import re

__all__ = ["RDate"]

QUARTER_FIRST_MTH = [1, 1, 1, 4, 4, 4, 7, 7, 7, 10, 10, 10]

SPLITTER = re.compile("([\+,\-]\d*\w+)")
OPERANDS = {"+", "-"}


###############################################################################
class RDate(object):
    """
    A date shift object that can be added to Dates to generate shifted dates.
    """
    __slots__ = ("rule", "calendar")

    # -------------------------------------------------------------------------
    def __init__(self, rule, calendar=None):
        """
        Inputs:
            rule     - a string specifying relative shift (see below for valid
                       rules).
            calendar - a holiday calendar name used to identify business days
        Rule definitions:
            d = add calendar day
            b = add business day
            w = add calendar week
            m = add calendar month
            y = add calendar year
            c = go to the required day in the month
            e = go to end of month (ignores num)
            J = go to first calendar day of month (ignores num)
            M = go to closest Monday as specified by num
            F = go to closest Friday as specified by num
            q = go to beginning of the quarter (ignores num)
            Q = go to end of the quarter (ignores num)
            A = go to beginning of the year (ignores num)
            E = go to end of the year (ignores num)
        """
        self.rule = rule
        self.calendar = calendar or HolidayCalendar()

    # -------------------------------------------------------------------------
    def apply_rule(self, d):
        # --- rule processing. If no operator is defined assume it's "+"
        if self.rule[0] in OPERANDS:
            atomic = SPLITTER.split(self.rule)[1::2]
        else:
            atomic = SPLITTER.split("+" + self.rule)[1::2]

        # --- iteratively apply each atomic rule
        for rule in atomic:
            op = rule[0:-1]
            r = rule[-1]
            if op in OPERANDS:
                op += "1"
            # --- look for the proper rule to apply
            if r == "d":
                d += relativedelta(days=int(op))
            elif r == "b":
                nb = int(op[1:])
                op1 = int(op[0] + "1")
                if nb == 0 and self.calendar.is_holiday(d):
                    # --- go to the next (or previous) business day only if
                    #     d is not already a business day
                    nb = 1
                for i in range(nb):
                    d += relativedelta(days=op1)
                    while self.calendar.is_holiday(d):
                        d += relativedelta(days=op1)
            elif r == "w":
                d += relativedelta(weeks=int(op))
            elif r == "m":
                d += relativedelta(months=int(op))
            elif r == "y":
                d += relativedelta(years=int(op))
            elif r == "c":
                d += relativedelta(day=int(op))
            elif r == "e":
                d += relativedelta(day=31)
            elif r == "J":
                d += relativedelta(day=1)
            elif r == "M":
                d += relativedelta(weekday=MO(int(op)))
            elif r == "F":
                d += relativedelta(weekday=FR(int(op)))
            elif r == "q":
                d = d.replace(day=1, month=QUARTER_FIRST_MTH[d.month-1])
            elif r == "Q":
                d = d.replace(day=1, month=QUARTER_FIRST_MTH[d.month-1]+2)
                d += relativedelta(day=31)
            elif r == "A":
                d = d.replace(day=1, month=1)
            elif r == "E":
                d = d.replace(day=31, month=12)
            else:
                raise NameError("Atomic rule {0:s} is unknown. "
                                "Full rule is {1:s}".format(r, rule))

        # --- conversion to Date is needed here because applying a
        #     relativedelta to a Date returns a datetime object
        return Date.parse(d)

    # -------------------------------------------------------------------------
    #  relative date algebra

    def __radd__(self, date):
        # --- check against the supercalss datetime.datetime
        if not isinstance(date, (datetime.date, datetime.datetime)):
            raise ValueError("RDate can only be applied to a Date. "
                             "{0!s} was passed instead".format(date.__class__))
        return self.apply_rule(date)
